# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from montepy.input_parser.tokens import MCNP_Lexer
from montepy.input_parser import syntax_node
from sly import Parser
import sly

_dec = sly.yacc._decorator


class MetaBuilder(sly.yacc.ParserMeta):
    """
    Custom MetaClass for allowing subclassing of MCNP_Parser.


    .. versionadded:: 0.2.0
        This was added with the major parser rework.

    Note: overloading functions is not allowed.
    """

    protected_names = {
        "debugfile",
        "errok",
        "error",
        "index_position",
        "line_position",
        "log",
        "parse",
        "restart",
        "tokens",
        "dont_copy",
    }

    def __new__(meta, classname, bases, attributes):
        if classname != "MCNP_Parser":
            for basis in bases:
                MetaBuilder._flatten_rules(classname, basis, attributes)
        cls = super().__new__(meta, classname, bases, attributes)
        return cls

    @staticmethod
    def _flatten_rules(classname, basis, attributes):
        for attr_name in dir(basis):
            if (
                not attr_name.startswith("_")
                and attr_name not in MetaBuilder.protected_names
                and attr_name not in attributes.get("dont_copy", set())
            ):
                func = getattr(basis, attr_name)
                attributes[attr_name] = func
        parent = basis.__bases__
        for par_basis in parent:
            if par_basis != Parser:
                return


class SLY_Supressor:
    """
    This is a fake logger meant to mostly make warnings dissapear.

    .. versionadded:: 0.2.0
        This was added with the major parser rework.
    """

    def __init__(self):
        self._parse_fail_queue = []

    def debug(self, msg, *args, **kwargs):
        pass

    info = debug

    warning = debug

    error = debug

    critical = debug

    def parse_error(self, msg, token=None, lineno=0, index=0):
        """
        Adds a SLY parsing error to the error queue for being dumped later.

        :param msg: The message to display.
        :type msg: str
        :param token: the token that caused the error if any.
        :type token: Token
        :param lineno: the current lineno of the error (from SLY not the file), if any.
        :type lineno: int
        """
        self._parse_fail_queue.append(
            {"message": msg, "token": token, "line": lineno, "index": index}
        )

    def clear_queue(self):
        """
        Clears the error queue and returns all errors.

        Returns a list of dictionaries. The dictionary has the keys: "message", "token", "line.

        :returns: A list of the errors since the queue was last cleared.
        :rtype: list
        """
        ret = self._parse_fail_queue
        self._parse_fail_queue = []
        return ret

    def __len__(self):
        return len(self._parse_fail_queue)


class MCNP_Parser(Parser, metaclass=MetaBuilder):
    """
    Base class for all MCNP parsers that provides basics.

    .. versionadded:: 0.2.0
        This was added with the major parser rework.
    """

    # Remove this if trying to see issues with parser
    log = SLY_Supressor()
    tokens = MCNP_Lexer.tokens
    debugfile = None

    def restart(self):
        """
        Clears internal state information about the current parse.

        Should be ran before a new object is parsed.
        """
        self.log.clear_queue()
        super().restart()

    def parse(self, token_generator, input=None):
        """
        Parses the token stream and returns a syntax tree.

        If the parsing fails None will be returned.
        The error queue can be retrieved from ``parser.log.clear_queue()``.

        :param token_generator: the token generator from ``lexer.tokenize``.
        :type token_generator: generator
        :param input: the input that is being lexed and parsed.
        :type input: Input
        :rtype: SyntaxNode
        """
        self._input = input

        # debug every time a token is taken
        def gen_wrapper():
            while True:
                token = next(token_generator, None)
                self._debug_parsing_error(token)
                yield token

        # change to using `gen_wrapper()` to debug
        tree = super().parse(token_generator)
        # treat any previous errors as being fatal even if it recovered.
        if len(self.log) > 0:
            return None
        return tree

    precedence = (("left", SPACE), ("left", TEXT))

    @_("NUMBER", "NUMBER padding")
    def number_phrase(self, p):
        """
        A non-zero number with or without padding.

        :returns: a float ValueNode
        :rtype: ValueNode
        """
        return self._flush_phrase(p, float)

    @_("NUMBER", "NUMBER padding")
    def identifier_phrase(self, p):
        """
        A non-zero number with or without padding converted to int.

        :returns: an int ValueNode
        :rtype: ValueNode
        """
        return self._flush_phrase(p, int)

    @_(
        "numerical_phrase",
        "shortcut_phrase",
        "number_sequence numerical_phrase",
        "number_sequence shortcut_phrase",
    )
    def number_sequence(self, p):
        """
        A list of numbers.

        :rtype: ListNode
        """
        if len(p) == 1:
            sequence = syntax_node.ListNode("number sequence")
            if type(p[0]) == syntax_node.ListNode:
                return p[0]
            sequence.append(p[0])
        else:
            sequence = p[0]
            if type(p[1]) == syntax_node.ListNode:
                for node in p[1].nodes:
                    sequence.append(node)
            else:
                sequence.append(p[1])
        return sequence

    @_("number_phrase", "null_phrase")
    def numerical_phrase(self, p):
        """
        Any number, including 0, with its padding.

        :returns: a float ValueNode
        :rtype: ValueNode
        """
        return p[0]

    @_("numerical_phrase", "shortcut_phrase")
    def shortcut_start(self, p):
        return p[0]

    @_(
        "shortcut_start NUM_REPEAT",
        "shortcut_start REPEAT",
        "shortcut_start NUM_MULTIPLY",
        "shortcut_start MULTIPLY",
        "shortcut_start NUM_INTERPOLATE padding number_phrase",
        "shortcut_start INTERPOLATE padding number_phrase",
        "shortcut_start NUM_LOG_INTERPOLATE padding number_phrase",
        "shortcut_start LOG_INTERPOLATE padding number_phrase",
        "NUM_JUMP",
        "JUMP",
    )
    def shortcut_sequence(self, p):
        """
        A shortcut (repeat, multiply, interpolate, or jump).

        :returns: the parsed shortcut.
        :rtype: ShortcutNode
        """
        short_cut = syntax_node.ShortcutNode(p)
        if isinstance(p[0], syntax_node.ShortcutNode):
            list_node = syntax_node.ListNode("next_shortcuts")
            list_node.append(p[0])
            list_node.append(short_cut)
            return list_node
        return short_cut

    @_("shortcut_sequence", "shortcut_sequence padding")
    def shortcut_phrase(self, p):
        """
        A complete shortcut, which should be used, and not shortcut_sequence.

        :returns: the parsed shortcut.
        :rtype: ShortcutNode
        """
        sequence = p.shortcut_sequence
        if len(p) == 2:
            sequence.end_padding = p.padding
        return sequence

    @_("NULL", "NULL padding")
    def null_phrase(self, p):
        """
        A zero number with or without its padding.

        :returns: a float ValueNode
        :rtype: ValueNode
        """
        return self._flush_phrase(p, float)

    @_("NULL", "NULL padding")
    def null_ident_phrase(self, p):
        """
        A zero number with or without its padding, for identification.

        :returns: an int ValueNode
        :rtype: ValueNode
        """
        return self._flush_phrase(p, int)

    @_("TEXT", "TEXT padding")
    def text_phrase(self, p):
        """
        A string with or without its padding.

        :returns: a str ValueNode.
        :rtype: ValueNode
        """
        return self._flush_phrase(p, str)

    def _flush_phrase(self, p, token_type):
        """
        Creates a ValueNode.
        """
        if len(p) > 1:
            padding = p[1]
        else:
            padding = None
        return syntax_node.ValueNode(p[0], token_type, padding)

    @_("SPACE", "DOLLAR_COMMENT", "COMMENT")
    def padding(self, p):
        """
        Anything that is not semantically significant: white space, and comments.

        :returns: All sequential padding.
        :rtype: PaddingNode
        """
        if hasattr(p, "DOLLAR_COMMENT") or hasattr(p, "COMMENT"):
            is_comment = True
        else:
            is_comment = False
        return syntax_node.PaddingNode(p[0], is_comment)

    @_("padding SPACE", "padding DOLLAR_COMMENT", "padding COMMENT", 'padding "&"')
    def padding(self, p):
        """
        Anything that is not semantically significant: white space, and comments.

        :returns: All sequential padding.
        :rtype: PaddingNode
        """
        if hasattr(p, "DOLLAR_COMMENT") or hasattr(p, "COMMENT"):
            is_comment = True
        else:
            is_comment = False
        p[0].append(p[1], is_comment)
        return p[0]

    @_("parameter", "parameters parameter")
    def parameters(self, p):
        """
        A list of the parameters (key, value pairs) for this input.

        :returns: all parameters
        :rtype: ParametersNode
        """
        if len(p) == 1:
            params = syntax_node.ParametersNode()
            param = p[0]
        else:
            params = p[0]
            param = p[1]
        params.append(param)
        return params

    @_(
        "classifier param_seperator number_sequence",
        "classifier param_seperator text_phrase",
    )
    def parameter(self, p):
        """
        A singular Key-value pair.

        :returns: the parameter.
        :rtype: SyntaxNode
        """
        return syntax_node.SyntaxNode(
            p.classifier.prefix.value,
            {"classifier": p.classifier, "seperator": p.param_seperator, "data": p[2]},
        )

    @_("file_atom", "file_name file_atom")
    def file_name(self, p):
        """
        A file name.

        :rtype: str
        """
        ret = p[0]
        if len(p) > 1:
            ret += p[1]
        return ret

    @_(
        "TEXT",
        "FILE_PATH",
        "NUMBER",
        "PARTICLE",
        "INTERPOLATE",
        "JUMP",
        "KEYWORD",
        "LOG_INTERPOLATE",
        "NULL",
        "REPEAT",
        "SURFACE_TYPE",
        "THERMAL_LAW",
        "ZAID",
        "NUMBER_WORD",
    )
    def file_atom(self, p):
        return p[0]

    @_("file_name", "file_name padding")
    def file_phrase(self, p):
        """
        A file name with or without its padding.

        :returns: a str ValueNode.
        :rtype: ValueNode
        """
        return self._flush_phrase(p, str)

    @_("padding", "equals_sign", "padding equals_sign")
    def param_seperator(self, p):
        """
        The seperation between a key and value for a parameter.

        :returns: a str ValueNode
        :rtype: ValueNode
        """
        padding = p[0]
        if len(p) > 1:
            padding += p[1]
        return padding

    @_('"="', '"=" padding')
    def equals_sign(self, p):
        """
        The seperation between a key and value for a parameter.

        :returns: a str ValueNode
        :rtype: ValueNode
        """
        padding = syntax_node.PaddingNode(p[0])
        if hasattr(p, "padding"):
            padding += p.padding
        return padding

    @_('":" part', 'particle_type "," part')
    def particle_type(self, p):
        if hasattr(p, "particle_type"):
            token = p.particle_type.token + "".join(list(p)[1:])
            particle_node = syntax_node.ParticleNode("data particles", token)
        else:
            particle_node = syntax_node.ParticleNode("data particles", "".join(list(p)))

        return particle_node

    @_("PARTICLE", "PARTICLE_SPECIAL")
    def part(self, p):
        return p[0]

    @_(
        "TEXT",
        "KEYWORD",
        "PARTICLE",
        "SOURCE_COMMENT",
        "TALLY_COMMENT",
    )
    def data_prefix(self, p):
        return syntax_node.ValueNode(p[0], str)

    @_(
        "modifier data_prefix",
        "data_prefix",
        "classifier NUMBER",
        "classifier particle_type",
    )
    def classifier(self, p):
        """
        The classifier of a data input.

        This represents the first word of the data input.
        E.g.: ``M4``, `IMP:N`, ``F104:p``

        :rtype: ClassifierNode
        """
        if hasattr(p, "classifier"):
            classifier = p.classifier
        else:
            classifier = syntax_node.ClassifierNode()

        if hasattr(p, "modifier"):
            classifier.modifier = syntax_node.ValueNode(p.modifier, str)
        if hasattr(p, "data_prefix"):
            classifier.prefix = p.data_prefix
        if hasattr(p, "NUMBER"):
            classifier.number = syntax_node.ValueNode(p.NUMBER, int)
        if hasattr(p, "particle_type"):
            classifier.particles = p.particle_type
        return classifier

    @_("classifier padding", "classifier")
    def classifier_phrase(self, p):
        """
        A classifier with its padding.

        :rtype: ClassifierNode
        """
        classifier = p.classifier
        if len(p) > 1:
            classifier.padding = p.padding
        return classifier

    @_('"*"', "PARTICLE_SPECIAL")
    def modifier(self, p):
        """
        A character that modifies a classifier, e.g., ``*TR``.

        :returns: the modifier
        :rtype: str
        """
        if hasattr(p, "PARTICLE_SPECIAL"):
            if p.PARTICLE_SPECIAL == "*":
                return "*"
        return p[0]

    def error(self, token):
        """
        Default error handling.

        Puts the data into a queue that can be pulled out later for one final clear debug.

        :param token: the token that broke the parsing rules.
        :type token: Token
        """
        if token:
            lineno = getattr(token, "lineno", 0)
            if self._input and self._input.lexer:
                lexer = self._input.lexer
                index = lexer.find_column(lexer.text, token)
            else:
                index = 0
            if lineno:
                self.log.parse_error(
                    f"sly: Syntax error at line {lineno}, token={token.type}\n",
                    token,
                    lineno,
                    index,
                )
            else:
                self.log.parse_error(
                    f"sly: Syntax error, token={token.type}", token, lineno
                )
        else:
            self.log.parse_error("sly: Parse error in input. EOF\n")

    def _debug_parsing_error(self, token):  # pragma: no cover
        """
        A function that should be called from error when debugging a parsing error.

        Call this from the method error. Also you will need the relevant debugfile to be set and saving the parser
        tables to file. e.g.,

        debugfile = 'parser.out'
        """
        print(f"********* New Parsing Error from: {type(self)} ************ ")
        print(f"Token: {token}")
        print(f"State: {self.state}, statestack: {self.statestack}")
        print(f"Symstack: {self.symstack}")
        print(f"Log length: {len(self.log)}")
        print()



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from unittest import TestCase
import pytest

import montepy

from montepy.data_inputs.element import Element
from montepy.data_inputs.isotope import Isotope
from montepy.data_inputs.material import Material
from montepy.data_inputs.material_component import MaterialComponent
from montepy.data_inputs.thermal_scattering import ThermalScatteringLaw
from montepy.errors import MalformedInputError, UnknownElement
from montepy.input_parser.block_type import BlockType
from montepy.input_parser.mcnp_input import Input


class testMaterialClass(TestCase):
    def test_material_parameter_parsing(self):
        for line in ["M20 1001.80c 1.0 gas=0", "M20 1001.80c 1.0 gas = 0 nlib = 00c"]:
            input = Input([line], BlockType.CELL)
            material = Material(input)

    def test_material_validator(self):
        material = Material()
        with self.assertRaises(montepy.errors.IllegalState):
            material.validate()
        with self.assertRaises(montepy.errors.IllegalState):
            material.format_for_mcnp_input((6, 2, 0))

    def test_material_setter(self):
        in_str = "M20 1001.80c 0.5 8016.80c 0.5"
        input_card = Input([in_str], BlockType.DATA)
        material = Material(input_card)
        material.number = 30
        self.assertEqual(material.number, 30)
        with self.assertRaises(TypeError):
            material.number = "foo"
        with self.assertRaises(ValueError):
            material.number = -5

    def test_material_str(self):
        in_str = "M20 1001.80c 0.5 8016.80c 0.4 94239.80c 0.1"
        input_card = Input([in_str], BlockType.DATA)
        material = Material(input_card)
        answers = """\
MATERIAL: 20 fractions: atom
 H-1     (80c) 0.5
 O-16    (80c) 0.4
Pu-239   (80c) 0.1
"""
        output = repr(material)
        print(output)
        assert output == answers
        output = str(material)
        assert output == "MATERIAL: 20, ['hydrogen', 'oxygen', 'plutonium']"

    def test_material_sort(self):
        in_str = "M20 1001.80c 0.5 8016.80c 0.5"
        input_card = Input([in_str], BlockType.DATA)
        material1 = Material(input_card)
        in_str = "M30 1001.80c 0.5 8016.80c 0.5"
        input_card = Input([in_str], BlockType.DATA)
        material2 = Material(input_card)
        sort_list = sorted([material2, material1])
        answers = [material1, material2]
        for i, mat in enumerate(sort_list):
            self.assertEqual(mat, answers[i])


def test_material_format_mcnp():
    in_strs = ["M20 1001.80c 0.5", "     8016.80c         0.5"]
    input_card = Input(in_strs, BlockType.DATA)
    material = Material(input_card)
    material.number = 25
    answers = ["M25 1001.80c 0.5", "     8016.80c         0.5"]
    output = material.format_for_mcnp_input((6, 2, 0))
    assert output == answers


@pytest.mark.parametrize(
    "isotope, conc, error",
    [
        ("1001.80c", -0.1, ValueError),
        ("1001.80c", "hi", TypeError),
        ("hi", 1.0, ValueError),
    ],
)
def test_material_comp_init(isotope, conc, error):
    with pytest.raises(error):
        MaterialComponent(Isotope(isotope, suppress_warning=True), conc, True)


def test_mat_comp_init_warn():
    with pytest.warns(DeprecationWarning):
        MaterialComponent(Isotope("1001.80c", suppress_warning=True), 0.1)


def test_material_comp_fraction_setter():
    comp = MaterialComponent(Isotope("1001.80c", suppress_warning=True), 0.1, True)
    comp.fraction = 5.0
    assert comp.fraction == pytest.approx(5.0)
    with pytest.raises(ValueError):
        comp.fraction = -1.0
    with pytest.raises(TypeError):
        comp.fraction = "hi"


def test_material_comp_fraction_str():
    comp = MaterialComponent(Isotope("1001.80c", suppress_warning=True), 0.1, True)
    str(comp)
    repr(comp)


def test_material_update_format():
    in_str = "M20 1001.80c 0.5 8016.80c 0.5"
    input_card = Input([in_str], BlockType.DATA)
    material = Material(input_card)
    assert material.format_for_mcnp_input((6, 2, 0)) == [in_str]
    material.number = 5
    print(material.format_for_mcnp_input((6, 2, 0)))
    assert "8016" in material.format_for_mcnp_input((6, 2, 0))[0]
    # addition
    isotope = Isotope("2004.80c", suppress_warning=True)
    with pytest.deprecated_call():
        material.material_components[isotope] = MaterialComponent(isotope, 0.1, True)
        print(material.format_for_mcnp_input((6, 2, 0)))
        assert "2004" in material.format_for_mcnp_input((6, 2, 0))[0]
        # update
        isotope = list(material.material_components.keys())[-1]
        print(material.material_components.keys())
        material.material_components[isotope].fraction = 0.7
        print(material.format_for_mcnp_input((6, 2, 0)))
        assert "0.7" in material.format_for_mcnp_input((6, 2, 0))[0]
        material.material_components[isotope] = MaterialComponent(isotope, 0.6, True)
        print(material.format_for_mcnp_input((6, 2, 0)))
        assert "0.6" in material.format_for_mcnp_input((6, 2, 0))[0]
        # delete
        del material.material_components[isotope]
        print(material.format_for_mcnp_input((6, 2, 0)))
        assert "8016" in material.format_for_mcnp_input((6, 2, 0))[0]


@pytest.mark.parametrize(
    "line, mat_number, is_atom, fractions",
    [
        ("M20 1001.80c 0.5 8016.710nc 0.5", 20, True, [0.5, 0.5]),
        ("m1 1001 0.33 8016 0.666667", 1, True, [0.33, 0.666667]),
        ("M20 1001 0.5 8016 0.5", 20, True, [0.5, 0.5]),
        ("M20 1001.80c -0.5 8016.80c -0.5", 20, False, [0.5, 0.5]),
        ("M20 1001.80c -0.5 8016.710nc -0.5", 20, False, [0.5, 0.5]),
        ("M20 1001.80c 0.5 8016.80c 0.5 Gas=1", 20, True, [0.5, 0.5]),
        (
            "m1      8016.71c  2.6999999-02 8017.71c  9.9999998-01 plib=84p",
            1,
            True,
            [2.6999999e-2, 9.9999998e-01],
        ),
        *[
            (f"M20 1001.80c 0.5 8016.80c 0.5 {part}={lib}", 20, True, [0.5, 0.5])
            for part, lib in [
                ("nlib", "80c"),
                ("nlib", "701nc"),
                ("estep", 1),
                ("pnlib", "710nc"),
                ("slib", "80c"),
            ]
        ],
    ],
)
def test_material_init(line, mat_number, is_atom, fractions):
    input = Input([line], BlockType.DATA)
    material = Material(input)
    assert material.number == mat_number
    assert material.old_number == mat_number
    assert material.is_atom_fraction == is_atom
    with pytest.deprecated_call():
        for component, gold in zip(material.material_components.values(), fractions):
            assert component.fraction == pytest.approx(gold)
    if "gas" in line:
        assert material.parameters["gas"]["data"][0].value == pytest.approx(1.0)


@pytest.mark.parametrize(
    "line", ["Mfoo", "M-20", "M20 1001.80c foo", "M20 1001.80c 0.5 8016.80c -0.5"]
)
def test_bad_init(line):
    # test invalid material number
    input = Input([line], BlockType.DATA)
    with pytest.raises(MalformedInputError):
        Material(input)


class TestIsotope(TestCase):
    def test_isotope_init(self):
        with pytest.warns(FutureWarning):
            isotope = Isotope("1001.80c")
        self.assertEqual(isotope.ZAID, "1001")
        self.assertEqual(isotope.Z, 1)
        self.assertEqual(isotope.A, 1)
        self.assertEqual(isotope.element.Z, 1)
        self.assertEqual(isotope.library, "80c")
        with self.assertRaises(ValueError):
            Isotope("1001.80c.5", suppress_warning=True)
        with self.assertRaises(ValueError):
            Isotope("hi.80c", suppress_warning=True)

    def test_isotope_metastable_init(self):
        isotope = Isotope("13426.02c", suppress_warning=True)
        self.assertEqual(isotope.ZAID, "13426")
        self.assertEqual(isotope.Z, 13)
        self.assertEqual(isotope.A, 26)
        self.assertTrue(isotope.is_metastable)
        self.assertEqual(isotope.meta_state, 1)
        isotope = Isotope("92635.02c", suppress_warning=True)
        self.assertEqual(isotope.A, 235)
        self.assertEqual(isotope.meta_state, 1)
        isotope = Isotope("92935.02c", suppress_warning=True)
        self.assertEqual(isotope.A, 235)
        self.assertEqual(isotope.meta_state, 4)
        self.assertEqual(isotope.mcnp_str(), "92935.02c")
        edge_cases = [
            ("4412", 4, 12, 1),
            ("4413", 4, 13, 1),
            ("4414", 4, 14, 1),
            ("36569", 36, 69, 2),
            ("77764", 77, 164, 3),
        ]
        for ZA, Z_ans, A_ans, isomer_ans in edge_cases:
            isotope = Isotope(ZA + ".80c", suppress_warning=True)
            self.assertEqual(isotope.Z, Z_ans)
            self.assertEqual(isotope.A, A_ans)
            self.assertEqual(isotope.meta_state, isomer_ans)
        with self.assertRaises(ValueError):
            isotope = Isotope("13826.02c", suppress_warning=True)

    def test_isotope_get_base_zaid(self):
        isotope = Isotope("92635.02c", suppress_warning=True)
        self.assertEqual(isotope.get_base_zaid(), 92235)

    def test_isotope_library_setter(self):
        isotope = Isotope("1001.80c", suppress_warning=True)
        isotope.library = "70c"
        self.assertEqual(isotope.library, "70c")
        with self.assertRaises(TypeError):
            isotope.library = 1

    def test_isotope_str(self):
        isotope = Isotope("1001.80c", suppress_warning=True)
        assert isotope.mcnp_str() == "1001.80c"
        assert isotope.nuclide_str() == "H-1.80c"
        assert repr(isotope) == "Isotope('H-1.80c')"
        assert str(isotope) == " H-1     (80c)"
        isotope = Isotope("94239.80c", suppress_warning=True)
        assert isotope.nuclide_str() == "Pu-239.80c"
        assert isotope.mcnp_str() == "94239.80c"
        assert repr(isotope) == "Isotope('Pu-239.80c')"
        isotope = Isotope("92635.80c", suppress_warning=True)
        assert isotope.nuclide_str() == "U-235m1.80c"
        assert isotope.mcnp_str() == "92635.80c"
        assert str(isotope) == " U-235m1 (80c)"
        assert repr(isotope) == "Isotope('U-235m1.80c')"
        # stupid legacy stupidity #486
        isotope = Isotope("95642", suppress_warning=True)
        assert isotope.nuclide_str() == "Am-242"
        assert isotope.mcnp_str() == "95642"
        assert repr(isotope) == "Isotope('Am-242')"
        isotope = Isotope("95242", suppress_warning=True)
        assert isotope.nuclide_str() == "Am-242m1"
        assert isotope.mcnp_str() == "95242"
        assert repr(isotope) == "Isotope('Am-242m1')"


class TestThermalScattering(TestCase):
    def test_thermal_scattering_init(self):
        # test wrong input type assertion
        input_card = Input(["M20"], BlockType.DATA)
        with self.assertRaises(MalformedInputError):
            ThermalScatteringLaw(input_card)

        input_card = Input(["Mt20 grph.20t"], BlockType.DATA)
        card = ThermalScatteringLaw(input_card)
        self.assertEqual(card.old_number, 20)
        self.assertEqual(card.thermal_scattering_laws, ["grph.20t"])

        input_card = Input(["Mtfoo"], BlockType.DATA)
        with self.assertRaises(MalformedInputError):
            ThermalScatteringLaw(input_card)
        input_card = Input(["Mt-20"], BlockType.DATA)
        with self.assertRaises(MalformedInputError):
            ThermalScatteringLaw(input_card)
        in_str = "M20 1001.80c 0.5 8016.80c 0.5"
        input_card = Input([in_str], BlockType.DATA)
        material = Material(input_card)
        card = ThermalScatteringLaw(material=material)
        self.assertEqual(card.parent_material, material)

    def test_thermal_scattering_particle_parser(self):
        # replicate issue #121
        input_card = Input(["Mt20 h-h2o.40t"], BlockType.DATA)
        card = ThermalScatteringLaw(input_card)
        self.assertEqual(card.old_number, 20)
        self.assertEqual(card.thermal_scattering_laws, ["h-h2o.40t"])

    def test_thermal_scatter_validate(self):
        thermal = ThermalScatteringLaw()
        with self.assertRaises(montepy.errors.IllegalState):
            thermal.validate()
        with self.assertRaises(montepy.errors.IllegalState):
            thermal.format_for_mcnp_input((6, 2, 0))
        material = Material()
        material.number = 1
        thermal._old_number = montepy.input_parser.syntax_node.ValueNode("1", int)
        thermal.update_pointers([material])
        with self.assertRaises(montepy.errors.IllegalState):
            thermal.validate()
        thermal._old_number = montepy.input_parser.syntax_node.ValueNode("2", int)
        with self.assertRaises(montepy.errors.MalformedInputError):
            thermal.update_pointers([material])

    def test_thermal_scattering_add(self):
        in_str = "Mt20 grph.20t"
        input_card = Input([in_str], BlockType.DATA)
        card = ThermalScatteringLaw(input_card)
        card.add_scattering_law("grph.21t")
        self.assertEqual(len(card.thermal_scattering_laws), 2)
        self.assertEqual(card.thermal_scattering_laws, ["grph.20t", "grph.21t"])
        card.thermal_scattering_laws = ["grph.22t"]
        self.assertEqual(card.thermal_scattering_laws, ["grph.22t"])

    def test_thermal_scattering_setter(self):
        in_str = "Mt20 grph.20t"
        input_card = Input([in_str], BlockType.DATA)
        card = ThermalScatteringLaw(input_card)
        laws = ["grph.21t"]
        card.thermal_scattering_laws = laws
        self.assertEqual(card.thermal_scattering_laws, laws)
        with self.assertRaises(TypeError):
            card.thermal_scattering_laws = 5
        with self.assertRaises(TypeError):
            card.thermal_scattering_laws = [5]

    def test_thermal_scattering_material_add(self):
        in_str = "M20 1001.80c 1.0"
        input_card = Input([in_str], BlockType.DATA)
        card = Material(input_card)
        card.add_thermal_scattering("grph.21t")
        self.assertEqual(len(card.thermal_scattering.thermal_scattering_laws), 1)
        self.assertEqual(card.thermal_scattering.thermal_scattering_laws, ["grph.21t"])
        card.thermal_scattering_laws = ["grph.22t"]
        self.assertEqual(card.thermal_scattering_laws, ["grph.22t"])
        with self.assertRaises(TypeError):
            card.add_thermal_scattering(5)

    def test_thermal_scattering_format_mcnp(self):
        in_str = "Mt20 grph.20t"
        input_card = Input([in_str], BlockType.DATA)
        card = ThermalScatteringLaw(input_card)
        in_str = "M20 1001.80c 0.5 8016.80c 0.5"
        input_card = Input([in_str], BlockType.DATA)
        material = Material(input_card)
        material.thermal_scattering = card
        card._parent_material = material
        material.thermal_scattering.thermal_scattering_laws = ["grph.20t"]
        self.assertEqual(card.format_for_mcnp_input((6, 2, 0)), ["Mt20 grph.20t "])

    def test_thermal_str(self):
        in_str = "Mt20 grph.20t"
        input_card = Input([in_str], BlockType.DATA)
        card = ThermalScatteringLaw(input_card)
        self.assertEqual(str(card), "THERMAL SCATTER: ['grph.20t']")
        self.assertEqual(
            repr(card),
            "THERMAL SCATTER: material: None, old_num: 20, scatter: ['grph.20t']",
        )


class TestElement(TestCase):
    def test_element_init(self):
        for Z in range(1, 119):
            element = Element(Z)
            self.assertEqual(element.Z, Z)
            # Test to ensure there are no missing elements
            name = element.name
            symbol = element.symbol

        with self.assertRaises(UnknownElement):
            Element(119)

        spot_check = {
            1: ("H", "hydrogen"),
            40: ("Zr", "zirconium"),
            92: ("U", "uranium"),
            94: ("Pu", "plutonium"),
            29: ("Cu", "copper"),
            13: ("Al", "aluminum"),
        }
        for z, (symbol, name) in spot_check.items():
            element = Element(z)
            self.assertEqual(z, element.Z)
            self.assertEqual(symbol, element.symbol)
            self.assertEqual(name, element.name)

    def test_element_str(self):
        element = Element(1)
        self.assertEqual(str(element), "hydrogen")
        self.assertEqual(repr(element), "Z=1, symbol=H, name=hydrogen")

    def test_get_by_symbol(self):
        element = Element.get_by_symbol("Hg")
        self.assertEqual(element.name, "mercury")
        with self.assertRaises(UnknownElement):
            Element.get_by_symbol("Hi")

    def test_get_by_name(self):
        element = Element.get_by_name("mercury")
        self.assertEqual(element.symbol, "Hg")
        with self.assertRaises(UnknownElement):
            Element.get_by_name("hudrogen")


class TestParticle(TestCase):
    def test_particle_str(self):
        part = montepy.Particle("N")
        self.assertEqual(str(part), "neutron")



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
__name__ = "montepy.data_inputs"
from .data_input import DataInput
from .material import Material
from .thermal_scattering import ThermalScatteringLaw
from .data_parser import parse_data



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from montepy.errors import *


class Element:
    """
    Class to represent an element e.g., Aluminum.

    :param Z: the Z number of the element
    :type Z: int
    :raises UnknownElement: if there is no element with that Z number.
    """

    def __init__(self, Z):
        self._Z = Z
        if Z not in self.__Z_TO_SYMBOL:
            raise UnknownElement(f"Z={Z}")

    @property
    def symbol(self):
        """
        The atomic symbol for this Element.

        :returns: the atomic symbol
        :rtype: str
        """
        return self.__Z_TO_SYMBOL[self.Z]

    @property
    def Z(self):
        """
        The atomic number for this Element.

        :returns: the atomic number
        :rtype: int
        """
        return self._Z

    @property
    def name(self):
        """
        The name of the element.

        :returns: the element's name.
        :rtype: str
        """
        return self.__ELEMENT_NAMES[self.symbol]

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"Z={self.Z}, symbol={self.symbol}, name={self.name}"

    def __hash__(self):
        return hash(self.Z)

    def __eq__(self, other):
        return self.Z == other.Z

    @classmethod
    def get_by_symbol(cls, symbol):
        """
        Get an element by it's symbol.

        E.g., get the element with Z=1 from "H".

        :returns: the element with this symbol
        :rtype: Element
        :raises UnknownElement: if there is no element with that symbol.
        """
        try:
            Z = cls.__SYMBOL_TO_Z[symbol]
            return cls(Z)
        except KeyError:
            raise UnknownElement(f"The symbol: {symbol}")

    @classmethod
    def get_by_name(cls, name):
        """
        Get an element by it's name.

        E.g., get the element with Z=1 from "hydrogen".

        :returns: the element with this name
        :rtype: Element
        :raises UnknownElement: if there is no element with that name.
        """
        try:
            symbol = cls.__NAMES_TO_SYMBOLS[name]
            return cls.get_by_symbol(symbol)
        except KeyError:
            raise UnknownElement(f"The name: {name}")

    __Z_TO_SYMBOL = {
        1: "H",
        2: "He",
        3: "Li",
        4: "Be",
        5: "B",
        6: "C",
        7: "N",
        8: "O",
        9: "F",
        10: "Ne",
        11: "Na",
        12: "Mg",
        13: "Al",
        14: "Si",
        15: "P",
        16: "S",
        17: "Cl",
        18: "Ar",
        19: "K",
        20: "Ca",
        21: "Sc",
        22: "Ti",
        23: "V",
        24: "Cr",
        25: "Mn",
        26: "Fe",
        27: "Co",
        28: "Ni",
        29: "Cu",
        30: "Zn",
        31: "Ga",
        32: "Ge",
        33: "As",
        34: "Se",
        35: "Br",
        36: "Kr",
        37: "Rb",
        38: "Sr",
        39: "Y",
        40: "Zr",
        41: "Nb",
        42: "Mo",
        43: "Tc",
        44: "Ru",
        45: "Rh",
        46: "Pd",
        47: "Ag",
        48: "Cd",
        49: "In",
        50: "Sn",
        51: "Sb",
        52: "Te",
        53: "I",
        54: "Xe",
        55: "Cs",
        56: "Ba",
        57: "La",
        58: "Ce",
        59: "Pr",
        60: "Nd",
        61: "Pm",
        62: "Sm",
        63: "Eu",
        64: "Gd",
        65: "Tb",
        66: "Dy",
        67: "Ho",
        68: "Er",
        69: "Tm",
        70: "Yb",
        71: "Lu",
        72: "Hf",
        73: "Ta",
        74: "W",
        75: "Re",
        76: "Os",
        77: "Ir",
        78: "Pt",
        79: "Au",
        80: "Hg",
        81: "Tl",
        82: "Pb",
        83: "Bi",
        84: "Po",
        85: "At",
        86: "Rn",
        87: "Fr",
        88: "Ra",
        89: "Ac",
        90: "Th",
        91: "Pa",
        92: "U",
        93: "Np",
        94: "Pu",
        95: "Am",
        96: "Cm",
        97: "Bk",
        98: "Cf",
        99: "Es",
        100: "Fm",
        101: "Md",
        102: "No",
        103: "Lr",
        104: "Rf",
        105: "Db",
        106: "Sg",
        107: "Bh",
        108: "Hs",
        109: "Mt",
        110: "Ds",
        111: "Rg",
        112: "Cn",
        113: "Nh",
        114: "Fl",
        115: "Mc",
        116: "Lv",
        117: "Ts",
        118: "Og",
    }
    __SYMBOL_TO_Z = {}
    for __Z, __symbol in __Z_TO_SYMBOL.items():
        __SYMBOL_TO_Z[__symbol] = __Z
    del __Z, __symbol
    __ELEMENT_NAMES = {
        "H": "hydrogen",
        "He": "helium",
        "Li": "lithium",
        "Be": "beryllium",
        "B": "boron",
        "C": "carbon",
        "N": "nitrogen",
        "O": "oxygen",
        "F": "fluorine",
        "Ne": "neon",
        "Na": "sodium",
        "Mg": "magnesium",
        "Al": "aluminum",
        "Si": "silicon",
        "P": "phosphorus",
        "S": "sulfur",
        "Cl": "chlorine",
        "Ar": "argon",
        "K": "potassium",
        "Ca": "calcium",
        "Sc": "scandium",
        "Ti": "titanium",
        "V": "vanadium",
        "Cr": "chromium",
        "Mn": "manganese",
        "Fe": "iron",
        "Co": "cobalt",
        "Ni": "nickel",
        "Cu": "copper",
        "Zn": "zinc",
        "Ga": "gallium",
        "Ge": "germanium",
        "As": "arsenic",
        "Se": "selenium",
        "Br": "bromine",
        "Kr": "krypton",
        "Rb": "rubidium",
        "Sr": "strontium",
        "Y": "yttrium",
        "Zr": "zirconium",
        "Nb": "niobium",
        "Mo": "molybdenum",
        "Tc": "technetium",
        "Ru": "ruthenium",
        "Rh": "rhodium",
        "Pd": "palladium",
        "Ag": "silver",
        "Cd": "cadmium",
        "In": "indium",
        "Sn": "tin",
        "Sb": "antimony",
        "Te": "tellurium",
        "I": "iodine",
        "Xe": "xenon",
        "Cs": "cesium",
        "Ba": "barium",
        "La": "lanthanum",
        "Ce": "cerium",
        "Pr": "praseodymium",
        "Nd": "neodymium",
        "Pm": "promethium",
        "Sm": "samarium",
        "Eu": "europium",
        "Gd": "gadolinium",
        "Tb": "terbium",
        "Dy": "dysprosium",
        "Ho": "holmium",
        "Er": "erbium",
        "Tm": "thulium",
        "Yb": "ytterbium",
        "Lu": "lutetium",
        "Hf": "hafnium",
        "Ta": "tantalum",
        "W": "tungsten",
        "Re": "rhenium",
        "Os": "osmium",
        "Ir": "iridium",
        "Pt": "platinum",
        "Au": "gold",
        "Hg": "mercury",
        "Tl": "thallium",
        "Pb": "lead",
        "Bi": "bismuth",
        "Po": "polonium",
        "At": "astatine",
        "Rn": "radon",
        "Fr": "francium",
        "Ra": "radium",
        "Ac": "actinium",
        "Th": "thorium",
        "Pa": "protactinium",
        "U": "uranium",
        "Np": "neptunium",
        "Pu": "plutonium",
        "Am": "americium",
        "Cm": "curium",
        "Bk": "berkelium",
        "Cf": "californium",
        "Es": "einsteinium",
        "Fm": "fermium",
        "Md": "mendelevium",
        "No": "nobelium",
        "Lr": "lawrencium",
        "Rf": "rutherfordium",
        "Db": "dubnium",
        "Sg": "seaborgium",
        "Bh": "bohrium",
        "Hs": "hassium",
        "Mt": "meitnerium",
        "Ds": "darmstradtium",
        "Rg": "roentgenium",
        "Cn": "copernicium",
        "Nh": "nihonium",
        "Fl": "flerovium",
        "Mc": "Moscovium",
        "Lv": "livermorium",
        "Ts": "tenessine",
        "Og": "oganesson",
    }

    __NAMES_TO_SYMBOLS = {}
    for __symbol, __name in __ELEMENT_NAMES.items():
        __NAMES_TO_SYMBOLS[__name] = __symbol
    del __symbol, __name



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import montepy
from montepy.errors import *
from montepy.geometry_operators import Operator
from montepy.input_parser.syntax_node import (
    GeometryTree,
    PaddingNode,
    ValueNode,
    CommentNode,
)
from montepy.utilities import *


class HalfSpace:
    """
    Class representing a geometry half_space.

    .. versionadded:: 0.2.0
        This was added as the core of the rework to how MCNP geometries are implemented.

    The term `half-spaces <https://en.wikipedia.org/wiki/Half-space_(geometry)>`_ in MontePy is used very loosely,
    and is not mathematically rigorous. In MontePy a divider is a something
    that splits a space (R\\ :sup:`3` ) into two half-spaces. At the simplest this would
    be a plane or other quadratic surface. There will always be two half-spaces,
    a negative, or inside (False) or positive, outside (True).
    This class proper is for binary trees implementing
    `constructive solid geometry (CSG) <https://en.wikipedia.org/wiki/Constructive_solid_geometry>`_ set logic.
    The same logic with half-spaces still apply as the intersection will always create two
    half-spaces (though one may be the empty set).
    In this case thinking of "inside" and "outside" may be more useful.

    HalfSpaces use binary operators to implement set logic.

    For Intersection:

    .. code-block:: python

        half_space = left & right
        half_space &= other_side

    For Union:

    .. code-block:: python

        half_space = left | right
        half_space |= other_side

    And you can also complement a HalfSpace:

    .. code-block:: python

        half_space = ~ other_half_space

    Parantheses are allowed, and handled properly

    .. code-block:: python

        half_space = +bottom & (-left | +right)

    :param left: The left side of the binary tree.
    :type left: HalfSpace
    :param operator: the operator to apply between the two branches.
    :type operator: Operator
    :param right: the right side of the binary tree.
    :type right: HalfSpace
    :param node: the node this was parsed from.
    :type node: GeometryTree
    """

    def __init__(self, left, operator, right=None, node=None):
        if not isinstance(left, HalfSpace):
            raise TypeError(f"left must be a HalfSpace. {left} given.")
        if not isinstance(right, (HalfSpace, type(None))):
            raise TypeError(f"right must be a HalfSpace, or None. {right} given.")
        if not isinstance(operator, Operator):
            raise TypeError(f"operator must be of type Operator. {operator} given.")
        if not isinstance(node, (GeometryTree, type(None))):
            raise TypeError(f"node must be a GeometryTree or None. {node} given.")
        if right is None and operator not in {Operator.COMPLEMENT, Operator.GROUP}:
            raise ValueError(f"Both sides required for: {operator}")
        self._left = left
        self._operator = operator
        self._right = right
        self._node = node
        self._cell = None

    @make_prop_pointer("_left", ())
    def left(self):
        """
        The left side of the binary tree of this half_space.

        :returns: the left side of the tree.
        :rtype: HalfSpace
        """
        pass

    @make_prop_pointer("_right", (), deletable=True)
    def right(self):
        """
        The right side of the binary tree of this half_space if any.

        :returns: the right side of the tree.
        :rtype: HalfSpace
        """
        pass

    @make_prop_pointer("_operator", Operator)
    def operator(self):
        """
        The operator for applying to this binary tree.

        :returns: the operator for the tree.
        :rtype: Operator
        """
        pass

    @make_prop_pointer("_node")
    def node(self):
        """
        The syntax node for this HalfSpace if any.

        If this was generated by :func:`parse_input_node`,
        that initial node will be given. If this was created from scratch,
        a new node will be generated prior as this is being written to file.

        :returns: the node for this tree.
        :rtype: GeometryTree
        """
        pass

    @staticmethod
    def parse_input_node(node):
        """
        Parses the given syntax node as a half_space.

        :param node: the Input syntax node to parse.
        :type node: GeometryTree
        :returns: the HalfSpace properly representing the input geometry.
        :rtype: HalfSpace
        """
        if not isinstance(node, GeometryTree):
            raise TypeError("Node must be a GeoemtryTree.")
        sides = []
        for side in (node.left, node.right):
            if side is None:
                continue
            if isinstance(side, ValueNode):
                is_cell = node.operator == Operator.COMPLEMENT
                sides.append(UnitHalfSpace.parse_input_node(side, is_cell))
            else:
                sides.append(HalfSpace.parse_input_node(side))
        if node.operator == Operator._SHIFT and isinstance(sides[0], UnitHalfSpace):
            return sides[0]
        if len(sides) == 1:
            sides.append(None)
        return HalfSpace(sides[0], node.operator, sides[1], node)

    def update_pointers(self, cells, surfaces, cell):
        """
        Update pointers, and link this object to other objects in the problem.

        This will:

        1. Link this HalfSpace (and its children) to the parent cell.
        2. Update the divider parameter to point to the relevant surface or cell.
        3. Update the parent's :func:`~montepy.cell.Cell.surfaces`, and :func:`~montepy.cell.Cell.complements`.

        :param cells: the cells in the problem.
        :type cells: Cells
        :param surfaces: The surfaces in the problem.
        :type surfaces: Surfaces
        :param cell: the cell this HalfSpace is tied to.
        :type cell: Cell
        """
        self._cell = cell
        self.left.update_pointers(cells, surfaces, cell)
        if self.right is not None:
            self.right.update_pointers(cells, surfaces, cell)

    def _add_new_children_to_cell(self, other):
        """
        Adds the cells and surfaces from a new tree to this parent cell.

        :param other: the other HalfSpace to work with.
        :type other: HalfSpace
        """
        if self._cell is None:
            return
        cells, surfaces = other._get_leaf_objects()
        for container, parent in zip(
            (cells, surfaces), (self._cell.complements, self._cell.surfaces)
        ):
            for item in container:
                if item not in parent:
                    parent.append(item)

    def remove_duplicate_surfaces(self, deleting_dict):
        """Updates old surface numbers to prepare for deleting surfaces.

        This will ensure any new surfaces or complements properly get added to the parent
        cell's :func:`~montepy.cell.Cell.surfaces` and :func:`~montepy.cell.Cell.complements`.

        :param deleting_dict: a dict of the surfaces to delete.
        :type deleting_dict: dict
        """
        _, surfaces = self._get_leaf_objects()
        new_deleting_dict = {}
        for dead_surface, new_surface in deleting_dict.items():
            if dead_surface in surfaces:
                new_deleting_dict[dead_surface] = new_surface
        if len(new_deleting_dict) > 0:
            self.left.remove_duplicate_surfaces(new_deleting_dict)
            if self.right is not None:
                self.right.remove_duplicate_surfaces(new_deleting_dict)

    def _get_leaf_objects(self):
        """
        Get all of the leaf objects for this tree.

        The leaf objects are the surfaces and cells used
        as dividers for this HalfSpace.

        :returns: a Tuple of two sets: cells, and surfaces.
        :rtype: Tuple
        """
        cells, surfaces = self.left._get_leaf_objects()
        if self.right:
            new_cells, new_surfaces = self.right._get_leaf_objects()
            cells |= new_cells
            surfaces |= new_surfaces
        return cells, surfaces

    def _update_values(self):
        self._ensure_has_nodes()
        self._update_node()
        if isinstance(self, UnitHalfSpace):
            return
        self.left._update_values()
        if self.right:
            self.right._update_values()

    def _ensure_has_nodes(self):
        """
        Ensures this HalfSpace and its children has the necessary syntax nodes.
        """
        self._ensure_has_parens()
        self.left._ensure_has_nodes()
        if self.right is not None:
            self.right._ensure_has_nodes()
        if self.node is None:
            if self.operator == Operator.INTERSECTION:
                operator = " "
            elif self.operator == Operator.UNION:
                operator = " : "
            elif self.operator == Operator.COMPLEMENT:
                operator = " #"
            elif self.operator == Operator.GROUP:
                operator = None
            operator = PaddingNode(operator)
            # Update trees
            if self.operator in {Operator.INTERSECTION, Operator.UNION}:
                ret = {"left": self.left.node, "operator": operator}
            elif self.operator == Operator.COMPLEMENT and isinstance(
                self.left, UnitHalfSpace
            ):
                ret = {"operator": operator, "left": self.left.node}
            else:
                ret = {
                    "operator": operator,
                    "start_pad": PaddingNode("("),
                    "left": self.left.node,
                    "end_pad": PaddingNode(")"),
                }
            # handle right side of tree
            if self.right is not None:
                ret["right"] = self.right.node
            self._node = GeometryTree(
                "default geometry", ret, self.operator.value, self.left, self.right
            )
        self.node.nodes["left"] = self.left.node
        if self.right is not None:
            self.node.nodes["right"] = self.right.node

    def _ensure_has_parens(self):
        """
        Ensures that when a parentheses is needed it is added.

        This detects unions below an intersection. It then "adds" a parentheses
        by adding a GROUP to the tree.

        This ignores parentheses needed for complements as its handled in _update_node.
        """
        if self.operator == Operator.INTERSECTION:
            if type(self) == type(self.left) and self.left.operator == Operator.UNION:
                self.left = HalfSpace(self.left, Operator.GROUP)
            if (
                self.right is not None
                and type(self) == type(self.right)
                and self.right.operator == Operator.UNION
            ):
                self.right = HalfSpace(self.right, Operator.GROUP)

    def _update_node(self):
        """
        Ensures that the syntax node properly reflects the current HalfSpace structure
        """
        try:
            operator_node = self.node.nodes["operator"]
            output = operator_node.format()
        except KeyError:
            output = ""
        if self.operator == Operator.INTERSECTION:
            if not output.isspace():
                self.__switch_operator(" ")
        elif self.operator == Operator.UNION:
            if ":" not in output:
                self.__switch_operator(":")
        elif self.operator == Operator.COMPLEMENT:
            if "#" not in output:
                self.__switch_operator("#")
                # change tree order
                self._node = GeometryTree(
                    "default geometry",
                    {
                        "operator": self.node.nodes["operator"],
                        "left": self.node.nodes["left"],
                    },
                    self.operator.value,
                    self.left.node,
                    None,
                )
        elif self.operator == Operator.GROUP:
            if "start_pad" not in self.node.nodes:
                self._node = GeometryTree(
                    "default Geometry",
                    {
                        "start_pad": PaddingNode("("),
                        "left": self.node.nodes["left"],
                        "end_pad": PaddingNode(")"),
                    },
                    self.operator.value,
                    self.left.node,
                    None,
                )

    def __switch_operator(self, new_symbol):
        """
        Updates the information about the operator in the syntax tree.
        """
        operator_node = self.node.nodes["operator"]
        operator_node._nodes = [
            (
                n.replace("#", " ").replace(":", " ")
                if not isinstance(n, CommentNode)
                else n
            )
            for n in operator_node.nodes
        ]
        if new_symbol == "#":
            operator_node._nodes[-1] = operator_node.nodes[-1][:-1] + "#"
        elif new_symbol == ":":
            output = operator_node.format()
            midpoint = len(output) // 2
            total_len = 0
            for i, node in enumerate(operator_node.nodes):
                if total_len + len(node) - 1 >= midpoint:
                    break
                total_len += len(node)
            index = midpoint - total_len
            operator_node._nodes[i] = node[:index] + ":" + node[index + 1 :]

    def __iand__(self, other):
        if not isinstance(other, HalfSpace):
            raise TypeError(f"Right hand side must be HalfSpace. {other} given.")
        if isinstance(self, UnitHalfSpace):
            return self & other
        right_leaf = (
            self.right is not None and isinstance(self.right, UnitHalfSpace)
        ) or self.right is None
        if right_leaf:
            if self.right is not None:
                self.right = self.right & other
                return self
            else:
                return (~self.left) & other
        self.right &= other
        self._add_new_children_to_cell(other)
        return self

    def __ior__(self, other):
        if not isinstance(other, HalfSpace):
            raise TypeError(f"Right hand side must be HalfSpace. {other} given.")
        if isinstance(self, UnitHalfSpace):
            return self | other
        right_leaf = (
            self.right is not None and isinstance(self.right, UnitHalfSpace)
        ) or self.right is None
        if right_leaf:
            if self.right is not None:
                self.right = self.right | other
                return self
            else:
                return (~self.left) | other
        self.right |= other
        self._add_new_children_to_cell(other)
        return self

    def __and__(self, other):
        if not isinstance(other, HalfSpace):
            raise TypeError(f"Right hand side must be HalfSpace. {other} given.")
        return HalfSpace(self, Operator.INTERSECTION, other)

    def __or__(self, other):
        if not isinstance(other, HalfSpace):
            raise TypeError(f"Right hand side must be HalfSpace. {other} given.")
        return HalfSpace(self, Operator.UNION, other)

    def __invert__(self):
        return HalfSpace(self, Operator.COMPLEMENT)

    def __len__(self):
        length = 0
        length += len(self.left)
        if self.right is not None:
            length += len(self.right)
        return length

    def __eq__(self, other):
        # don't allow subclassing on right side
        if type(self) != type(other):
            raise TypeError(f"HalfSpaces can only be compared to each other")
        if self.operator != other.operator:
            return False
        if (self.right is None) != (other.right is None):
            return False

        if self.left != other.left:
            return False
        if self.right is not None and self.right == other.right:
            return True
        elif self.right is None:
            return True
        return False

    def _generate_default_tree(self):
        pass

    def __str__(self):
        if self.operator == Operator.COMPLEMENT:
            return f"{self.operator.value}{self.left}"
        if self.operator == Operator.GROUP:
            return str(self.left)
        return f"({self.left}{self.operator.value}{self.right})"

    def __repr__(self):
        return f"(l: {repr(self.left)}| op: {self.operator}| r:{repr(self.right)})"


class UnitHalfSpace(HalfSpace):
    """
    The leaf node for the HalfSpace tree.
    
    .. versionadded:: 0.2.0
        This was added as the core of the rework to how MCNP geometries are implemented.

    This can only be used as leaves and represents one half_space of a 
    a divider.
    The easiest way to generate one is with the divider with unary operators. 
    For surfaces you can choose the positive (True) or negative (False) side quickly:

    .. code-block:: python

        bottom_half = -surf
        top_half = +surf

    For a cell you can only take the complement:
    
    .. code-block:: python
    
        comp = ~cell

    .. Note::
        
        When you complement a cell you don't actually get a UnitHalfSpace directly.
        You get a UnitHalfSpace wrapped with a complementing HalfSpace Tree

        .. code-block:: text

                     HalfSpace
                   /      #   \\
            UnitHalfSpace      None
                   |
                   Cell

    :param divider: the divider object
    :type divider: int, Cell, Surface
    :param side: which side the divider is on. For Cells this will be True.
    :type side: bool
    :param is_cell: Whether or not this is a cell or not for the divider.
    :type is_cell: bool
    :param node: the node if any this UnitHalfSpace was built from
    :type node: ValueNode
    """

    def __init__(self, divider, side, is_cell, node=None):
        if not isinstance(
            divider, (int, montepy.Cell, montepy.surfaces.surface.Surface)
        ):
            raise TypeError(
                f"divider must be an int, Cell, or Surface. {divider} given"
            )
        if not isinstance(side, bool):
            raise TypeError(f"side must be bool. {side} given.")
        if not isinstance(is_cell, bool):
            raise TypeError(f"is_cell must be bool. {is_cell} given.")
        if not isinstance(node, (ValueNode, type(None))):
            raise TypeError(f"node must be a ValueNode or None. {node} given.")
        self._divider = divider
        self._side = side
        self._is_cell = is_cell
        self._node = node
        self._cell = None

    @property
    def divider(self):
        """
        The divider this UnitHalfSpace is based on.

        :returns: the divider defining this HalfSpace
        :rtype: int, Cell, Surface
        """
        return self._divider

    # done manually to avoid circular imports
    @divider.setter
    def divider(self, div):
        if not isinstance(div, (montepy.Cell, montepy.surfaces.surface.Surface)):
            raise TypeError("Divider must be a Cell or Surface")
        if self.is_cell != isinstance(div, montepy.Cell):
            raise TypeError("Divider type must match with is_cell")
        self._divider = div
        if self._cell is not None:
            if self.is_cell:
                container = self._cell.complements
            else:
                container = self._cell.surfaces
            if div not in container:
                container.append(div)

    @make_prop_pointer("_is_cell", bool)
    def is_cell(self):
        """
        Whether or not the divider this uses is a cell.

        :returns: True if this is a cell based HalfSpace
        :rtype: bool
        """
        pass

    def __str__(self):
        side = ""
        if self.side:
            if not self.is_cell:
                side = "+"
        else:
            side = "-"
        if isinstance(self.divider, int):
            div = self.divider
        else:
            div = self.divider.number
        return f"{side}{div}"

    def __repr__(self):
        return f"s: {self.side}, is_c: {self.is_cell}, div: {self.divider}"

    @property
    def node(self):
        """
        The node that this UnitHalfSpace is based on if any.

        :returns: The ValueNode that this UnitHalfSpace is tied to.
        :rtype: ValueNode
        """
        return self._node

    @make_prop_pointer("_side", bool)
    def side(self):
        """
        Which side of the divider this HalfSpace is on.

        This maps the conventional positive/negative half_spaces of MCNP
        to boolean values. True is the positive side, and False is the negative one.
        For cells this is always True as the Complementing logic is handled by the parent
        binary tree.

        :returns: the side of the divider for the HalfSpace
        :rtype: bool
        """
        # make cells always "+"
        return self.is_cell or self._side

    @staticmethod
    def parse_input_node(node, is_cell=False):
        """
        Parses the given syntax node as a UnitHalfSpace.

        :param node: the Input syntax node to parse.
        :type node: ValueNode
        :param is_cell: Whether or not this UnitHalfSpace represents a cell.
        :type is_cell: bool
        :returns: the HalfSpace properly representing the input geometry.
        :rtype: UnitHalfSpace
        """
        if not isinstance(node, ValueNode):
            raise TypeError(f"Must be called on a ValueNode. {node} given.")
        if not isinstance(is_cell, bool):
            raise TypeError("is_cell must be a bool.")
        node.is_negatable_identifier = True
        if is_cell:
            side = True
        else:
            side = not node.is_negative
        return UnitHalfSpace(node.value, side, is_cell, node)

    def update_pointers(self, cells, surfaces, cell):
        self._cell = cell
        container = surfaces
        par_container = self._cell.surfaces
        if self._is_cell:
            container = cells
            par_container = self._cell.complements
        if isinstance(self.divider, int):
            try:
                self._divider = container[self._divider]
                if self._divider not in par_container:
                    par_container.append(self._divider)
            except KeyError:
                if self._is_cell:
                    raise BrokenObjectLinkError(
                        "Cell", self._cell.number, "Complement", self._divider
                    )
                raise BrokenObjectLinkError(
                    "Cell", self._cell.number, "Surface", self._divider
                )

    def _ensure_has_nodes(self):
        if self.node is None:
            if isinstance(self.divider, int):
                num = self.divider
            else:
                num = self.divider.number
            node = ValueNode(None, int, never_pad=True)
            node.value = num
            node.is_negatable_identifier = True
            node.is_negative = not self.side
            self._node = node

    def _update_node(self):
        if isinstance(self.divider, int):
            self._node.value = self.divider
        else:
            self._node.value = self.divider.number
        self._node.is_negative = not self.side

    def _get_leaf_objects(self):
        if self._is_cell:
            return ({self._divider}, set())
        return (set(), {self._divider})

    def remove_duplicate_surfaces(self, deleting_dict):
        """Updates old surface numbers to prepare for deleting surfaces.

        :param deleting_dict: a dict of the surfaces to delete.
        :type deleting_dict: dict
        """
        if not self.is_cell:
            if self.divider in deleting_dict:
                new_surface = deleting_dict[self.divider]
                self.divider = new_surface

    def __len__(self):
        return 1

    def __eq__(self, other):
        if not isinstance(other, UnitHalfSpace):
            raise TypeError("UnitHalfSpace can't be equal to other type")
        return (
            self.is_cell == other.is_cell
            and self.divider is other.divider
            and self.side == other.side
        )



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
from .block_type import BlockType
from collections import deque
from .. import errors
import itertools
import io
from montepy.constants import *
from montepy.errors import *
from montepy.input_parser.input_file import MCNP_InputFile
from montepy.input_parser.mcnp_input import Input, Message, ReadInput, Title
from montepy.input_parser.read_parser import ReadParser
from montepy.utilities import is_comment
import os
import warnings

reading_queue = []


def read_input_syntax(input_file, mcnp_version=DEFAULT_VERSION, replace=True):
    """
    Creates a generator function to return a new MCNP input for
    every new one that is encountered.

    This is meant to just handle the MCNP input syntax, it does not
    semantically parse the inputs.

    The version must be a three component tuple e.g., (6, 2, 0) and (5, 1, 60).


    :param input_file: the path to the input file to be read
    :type input_file: MCNP_InputFile
    :param mcnp_version: The version of MCNP that the input is intended for.
    :type mcnp_version: tuple
    :param replace: replace all non-ASCII characters with a space (0x20)
    :type replace: bool
    :returns: a generator of MCNP_Object objects
    :rtype: generator
    """
    global reading_queue
    reading_queue = deque()
    with input_file.open("r", replace=replace) as fh:
        yield from read_front_matters(fh, mcnp_version)
        yield from read_data(fh, mcnp_version)


def read_front_matters(fh, mcnp_version):
    """
    Reads the beginning of an MCNP file for all of the unusual data there.

    This is a generator function that will yield multiple :class:`MCNP_Input` instances.

    .. warning::
        This function will move the file handle forward in state.

    .. warning::
        This function will not close the file handle.

    .. versionchanged:: 0.2.0
        ``fh`` was changed to be an MCNP_InputFile to hold more information.

    :param fh: The file handle of the input file.
    :type fh: MCNP_InputFile
    :param mcnp_version: The version of MCNP that the input is intended for.
    :type mcnp_version: tuple

    :return: an instance of the Title class, and possible an instance of a Message class
    :rtype: MCNP_Object
    """
    is_in_message_block = False
    found_title = False
    lines = []
    raw_lines = []
    for i, line in enumerate(fh):
        if i == 0 and line.upper().startswith("MESSAGE:"):
            is_in_message_block = True
            raw_lines.append(line.rstrip())
            lines.append(line[9:])  # removes "MESSAGE: "
        elif is_in_message_block:
            if line.strip():
                raw_lines.append(line.rstrip())
                lines.append(line)
            # message block is terminated by a blank line
            else:
                yield Message(raw_lines, lines)
                is_in_message_block = False
        # title always follows complete message, or is first
        else:
            yield Title([line], line)
            break


def read_data(fh, mcnp_version, block_type=None, recursion=False):
    """
    Reads the bulk of an MCNP file for all of the MCNP data.

    This is a generator function that will yield multiple :class:`MCNP_Input` instances.

    .. warning::
        This function will move the file handle forward in state.

    .. warning::
        This function will not close the file handle.

    .. versionchanged:: 0.2.0
        ``file_wrapper`` was added to better track which file is being read.

    :param fh: The file handle of the input file.
    :type fh: MCNP_InputFile
    :param mcnp_version: The version of MCNP that the input is intended for.
    :type mcnp_version: tuple
    :param block_type: The type of block this file is in. This is only used with partial files read using the ReadInput.
    :type block_type: BlockType
    :param recursion: Whether or not this is being called recursively. If True this has been called
                         from read_data. This prevents the reading queue causing infinite recursion.
    :type recursion: bool

    :return: MCNP_Input instances: Inputs that represent the data in the MCNP input.
    :rtype: MCNP_Input

    """
    current_file = fh
    line_length = get_max_line_length(mcnp_version)
    block_counter = 0
    if block_type is None:
        block_type = BlockType.CELL
    continue_input = False
    has_non_comments = False
    input_raw_lines = []

    def flush_block():
        nonlocal block_counter, block_type
        if len(input_raw_lines) > 0:
            yield from flush_input()
        block_counter += 1
        if block_counter < 3:
            block_type = BlockType(block_counter)

    def flush_input():
        nonlocal input_raw_lines
        start_line = current_file.lineno + 1 - len(input_raw_lines)
        input = Input(
            input_raw_lines,
            block_type,
            current_file,
            start_line,
        )
        try:
            read_input = ReadInput(
                input_raw_lines, block_type, current_file, start_line
            )
            reading_queue.append((block_type, read_input.file_name, current_file.path))
            yield None
        except ValueError as e:
            if isinstance(e, ParsingError):
                raise e
            yield input
        continue_input = False
        input_raw_lines = []

    for line in fh:
        line = line.expandtabs(TABSIZE)
        line_is_comment = is_comment(line)
        # transition to next block with blank line
        if not line.strip():
            yield from flush_block()
            has_non_comments = False
            continue
        # if a new input
        if (
            line[0:BLANK_SPACE_CONTINUE].strip()
            and not continue_input
            and not line_is_comment
            and has_non_comments
            and input_raw_lines
        ):
            yield from flush_input()
        # die if it is a vertical syntax format
        if "#" in line[0:BLANK_SPACE_CONTINUE] and not line_is_comment:
            raise errors.UnsupportedFeature("Vertical Input format is not allowed")
        # cut line down to allowed length
        old_line = line
        line = line[:line_length]
        if len(old_line) != len(line):
            warnings.warn(
                f"The line: {old_line} exceeded the allowed line length of: {line_length} for MCNP {mcnp_version}",
                errors.LineOverRunWarning,
            )
        if line.endswith(" &\n"):
            continue_input = True
        else:
            continue_input = False
        has_non_comments = has_non_comments or not line_is_comment
        input_raw_lines.append(line.rstrip())
    yield from flush_block()

    if not recursion:
        path = os.path.dirname(fh.name)
        while reading_queue:
            block_type, file_name, parent = reading_queue.popleft()
            new_wrapper = MCNP_InputFile(os.path.join(path, file_name), parent)
            with new_wrapper.open("r") as sub_fh:
                new_wrapper = MCNP_InputFile(file_name, parent)
                for input in read_data(sub_fh, mcnp_version, block_type, True):
                    yield input



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
import copy
from montepy.cells import Cells
from montepy.constants import BLANK_SPACE_CONTINUE
from montepy.data_inputs import importance, fill, lattice_input, universe_input, volume
from montepy.data_inputs.data_parser import PREFIX_MATCHES
from montepy.input_parser.cell_parser import CellParser
from montepy.input_parser import syntax_node
from montepy.errors import *
from montepy.numbered_mcnp_object import Numbered_MCNP_Object
from montepy.data_inputs.material import Material
from montepy.geometry_operators import Operator
from montepy.surfaces.half_space import HalfSpace, UnitHalfSpace
from montepy.surfaces.surface import Surface
from montepy.surface_collection import Surfaces
from montepy.universe import Universe
from montepy.utilities import *
import numbers


def _number_validator(self, number):
    if number <= 0:
        raise ValueError("number must be > 0")
    if self._problem:
        self._problem.cells.check_number(number)


def _link_geometry_to_cell(self, geom):
    geom._cell = self
    geom._add_new_children_to_cell(geom)


class Cell(Numbered_MCNP_Object):
    """
    Object to represent a single MCNP cell defined in CSG.

    .. versionchanged:: 0.2.0
        Removed the ``comments`` argument due to overall simplification of init process.


    :param input: the input for the cell definition
    :type input: Input

    .. seealso::

            * :manual63sec:`5.2`
            * :manual62:`55`
    """

    _ALLOWED_KEYWORDS = {
        "IMP",
        "VOL",
        "PWT",
        "EXT",
        "FCL",
        "WWN",
        "DXC",
        "NONU",
        "PD",
        "TMP",
        "U",
        "TRCL",
        "LAT",
        "FILL",
        "ELPT",
        "COSY",
        "BFLCL",
        "UNC",
    }
    _INPUTS_TO_PROPERTY = {
        importance.Importance: ("_importance", False),
        volume.Volume: ("_volume", True),
        universe_input.UniverseInput: ("_universe", True),
        lattice_input.LatticeInput: ("_lattice", True),
        fill.Fill: ("_fill", True),
    }
    _parser = CellParser()

    def __init__(self, input=None):
        self._material = None
        self._old_number = self._generate_default_node(int, -1)
        self._load_blank_modifiers()
        self._old_mat_number = self._generate_default_node(int, -1)
        self._density_node = self._generate_default_node(float, None)
        self._surfaces = Surfaces()
        self._complements = Cells()
        self._number = self._generate_default_node(int, -1)
        super().__init__(input, self._parser)
        if not input:
            self._generate_default_tree()
        self._old_number = copy.deepcopy(self._tree["cell_num"])
        self._number = self._tree["cell_num"]
        mat_tree = self._tree["material"]
        self._old_mat_number = mat_tree["mat_number"]
        self._density_node = mat_tree["density"]
        self._density_node.is_negatable_float = True
        if self.old_mat_number != 0:
            self._is_atom_dens = not self._density_node.is_negative
        self._parse_geometry()
        self._parse_keyword_modifiers()

    def _parse_geometry(self):
        """
        Parses the cell's geometry definition, and stores it
        """
        geometry = self._tree["geometry"]
        if geometry is not None:
            self._geometry = HalfSpace.parse_input_node(geometry)
        else:
            self._geometry = None

    def _parse_keyword_modifiers(self):
        """
        Parses the parameters to make the object and load as an attribute
        """
        found_class_prefixes = set()
        for key, value in self.parameters.nodes.items():
            for input_class in PREFIX_MATCHES:
                prefix = input_class._class_prefix()
                if input_class in Cell._INPUTS_TO_PROPERTY and prefix in key.lower():
                    attr, ban_repeat = Cell._INPUTS_TO_PROPERTY[input_class]
                    key = str(value["classifier"]).lower()
                    found_class_prefixes.add(value["classifier"].prefix.value.lower())
                    input = input_class(in_cell_block=True, key=key, value=value)
                    if not getattr(self, attr).set_in_cell_block:
                        setattr(self, attr, input)
                    else:
                        if not ban_repeat:
                            getattr(self, attr).merge(
                                input_class(in_cell_block=True, key=key, value=value)
                            )
        # Add defaults to tree
        for input_class, (attr, _) in self._INPUTS_TO_PROPERTY.items():
            has_imp = False
            class_pref = input_class._class_prefix()
            if class_pref in found_class_prefixes:
                continue
            if class_pref == "imp":
                for key in self._tree["parameters"].nodes.keys():
                    if class_pref in key:
                        has_imp = True
                        break
            if (class_pref == "imp" and not has_imp) or class_pref != "imp":
                tree = getattr(self, attr)._tree
                self._tree["parameters"].append(tree, True)

    def _load_blank_modifiers(self):
        """
        Goes through and populates all the modifier attributes
        """
        for input_class, (attr, _) in self._INPUTS_TO_PROPERTY.items():
            setattr(self, attr, input_class(in_cell_block=True))

    @property
    def importance(self):
        """
        The importances for this cell for various particle types.

        Each particle's importance is a property of Importance.
        e.g., ``cell.importance.photon = 1.0``.

        :returns: the importance for the Cell.
        :rtype: Importance
        """
        return self._importance

    @property
    def universe(self):
        """
        The Universe that this cell is in.

        :returns: the Universe the cell is in.
        :rtype: Universe
        """
        return self._universe.universe

    @property
    def fill(self):
        """
        the Fill object representing how this cell is filled.

        This not only describes the universe that is filling this,
        but more complex things like transformations, and matrix fills.

        :returns: The Fill object of how this cell is to be filled.
        :rtype: Fill
        """
        return self._fill

    @universe.setter
    def universe(self, value):
        if not isinstance(value, Universe):
            raise TypeError("universe must be set to a Universe")
        self._universe.universe = value

    @property
    def not_truncated(self):
        """
        Indicates if this cell has been marked as not being truncated for optimization.

        See Note 1 from section 3.3.1.5.1 of the user manual (LA-UR-17-29981).

        Note this can be set to True iff that this cell is not in Universe 0.

            Note 1. A problem will run faster by preceding the U card entry with a minus sign for any
            cell that is not truncated by the boundary of any higher-level cell. (The minus sign indicates
            that calculating distances to boundary in higher-level cells can be omitted.) Use this
            capability with EXTREME CAUTION; MCNP6 cannot detect errors in this feature because
            the logic that enables detection is omitted by the presence of the negative universe. Extremely
            wrong answers can be quietly calculated. Plot several views of the geometry or run with the
            VOID card to check for errors.

            -- LA-UR-17-29981.

        :rtype: bool
        :returns: True if this cell has been marked as not being truncated by the parent filled cell.
        """
        if self.universe.number == 0:
            return False
        return self._universe.not_truncated

    @not_truncated.setter
    def not_truncated(self, value):
        if not isinstance(value, bool):
            raise TypeError("not_truncated_by_parent must be a bool")
        if self.universe.number == 0 and value:
            raise ValueError("can't specify if cell is truncated for universe 0")
        self._universe._not_truncated = value

    @property
    def old_universe_number(self):
        """
        The original universe number read in from the input file.

        :returns: the number of the Universe for the cell in the input file.
        :rtype: int
        """
        return self._universe.old_number

    @property
    def lattice(self):
        """
        The type of lattice being used by the cell.

        :returns: the type of lattice being used
        :rtype: Lattice
        """
        return self._lattice.lattice

    @lattice.setter
    def lattice(self, value):
        self._lattice.lattice = value

    @lattice.deleter
    def lattice(self):
        self._lattice.lattice = None

    @property
    def volume(self):
        """
        The volume for the cell.

        Will only return a number if the volume has been manually set.

        :returns: the volume that has been manually set or None.
        :rtype: float, None
        """
        return self._volume.volume

    @volume.setter
    def volume(self, value):
        self._volume.volume = value

    @volume.deleter
    def volume(self):
        del self._volume.volume

    @property
    def volume_mcnp_calc(self):
        """
        Indicates whether or not MCNP will attempt to calculate the cell volume.

        This can be disabled by either manually setting the volume or disabling
        this calculation globally.
        This does not guarantee that MCNP will able to calculate the volume.
        Complex geometries may make this impossible.

        See :func:`~montepy.cells.Cells.allow_mcnp_volume_calc`

        :returns: True iff MCNP will try to calculate the volume for this cell.
        :rtype: bool
        """
        return self._volume.is_mcnp_calculated

    @property
    def volume_is_set(self):
        """
        Whether or not the volume for this cell has been set.

        :returns: true if the volume is manually set.
        :rtype: bool
        """
        return self._volume.set

    @make_prop_val_node("_old_number")
    def old_number(self):
        """
        The original cell number provided in the input file

        :rtype: int
        """
        pass

    @make_prop_val_node("_number", int, validator=_number_validator)
    def number(self):
        """
        The current cell number that will be written out to a new input.

        :rtype: int
        """
        pass

    @make_prop_pointer("_material", (Material, type(None)), deletable=True)
    def material(self):
        """
        The Material object for the cell.

        If the material is None this is considered to be voided.

        :rtype: Material
        """
        pass

    @make_prop_pointer("_geometry", HalfSpace, validator=_link_geometry_to_cell)
    def geometry(self):
        """
        The Geometry for this problem.

        .. versionadded:: 0.2.0
            Added with the new ability to represent true CSG geometry logic.

        The HalfSpace tree that is able to represent this cell's geometry.
        MontePy's geometry is based upon dividers, which includes both Surfaces, and cells.
        A half-space is created by choosing one side of the divider.
        A divider will always create two half-spaces; only one of which can be finite.

        For instance a plane creates two infinite half-spaces, one above and one below.
        A finite cell also creates two half-spaces; inside, and outside.

        These halfspaces can then be combined with set-logic to make a new half-space.

        To generate a halfspace from a surface you must specify the positive or negative side:

        .. code-block:: python

            half_space = +plane

        To complement a cell you must invert it:

        .. code-block:: python

           half_space = ~cell

        To create more complex geometry you can use binary and ``&`` as an intersection, and binary or ``|`` as a
        union:

        .. code-block:: python

            half_space = -cylinder & + bottom & - top

        For better documentation please refer to `OpenMC
        <https://docs.openmc.org/en/stable/usersguide/geometry.html>`_.

        :returns: this cell's geometry
        :rtype: HalfSpace
        """
        pass

    @property
    def geometry_logic_string(self):  # pragma: no cover
        """
        The original geoemtry input string for the cell.

        .. warning::
            .. deprecated:: 0.2.0
                This was removed to allow for :func:`geometry` to truly implement CSG geometry.

        :raise DeprecationWarning: Will always be raised as an error (which will cause program to halt).
        """
        raise DeprecationWarning(
            "Geometry_logic_string has been removed from cell. Use Cell.geometry instead."
        )

    @make_prop_val_node(
        "_density_node", (float, int, type(None)), base_type=float, deletable=True
    )
    def _density(self):
        """
        This is a wrapper to allow using the prop_val_node with mass_density and atom_density.
        """
        pass

    @property
    def atom_density(self) -> float:
        """
        The atom density of the material in the cell, in a/b-cm.

        :returns: the atom density. If no density is set or it is in mass density will return None.
        :rtype: float, None
        """
        if self._density and not self._is_atom_dens:
            raise AttributeError(f"Cell {self.number} is in mass density.")
        return self._density

    @atom_density.setter
    def atom_density(self, density: float):
        if not isinstance(density, numbers.Number):
            raise TypeError("Atom density must be a number.")
        elif density < 0:
            raise ValueError("Atom density must be a positive number.")
        self._is_atom_dens = True
        self._density = float(density)

    @atom_density.deleter
    def atom_density(self):
        self._density = None

    @property
    def mass_density(self) -> float:
        """
        The mass density of the material in the cell, in g/cc.

        :returns: the mass density. If no density is set or it is in atom density will return None.
        :rtype: float, None
        """
        if self._density and self._is_atom_dens:
            raise AttributeError(f"Cell {self.number} is in atom density.")
        return self._density

    @mass_density.setter
    def mass_density(self, density: float):
        if not isinstance(density, numbers.Number):
            raise TypeError("Mass density must be a number.")
        elif density < 0:
            raise ValueError("Mass density must be a positive number.")
        self._is_atom_dens = False
        self._density = float(density)

    @mass_density.deleter
    def mass_density(self):
        self._density = None

    @property
    def is_atom_dens(self):
        """
        Whether or not the density is in atom density [a/b-cm].

        True means it is in atom density, False means mass density [g/cc].

        :rtype: bool
        """
        return self._is_atom_dens

    @make_prop_val_node("_old_mat_number")
    def old_mat_number(self):
        """
        The material number provided in the original input file

        :rtype: int
        """
        pass

    @make_prop_pointer("_surfaces")
    def surfaces(self):
        """
        List of the Surface objects associated with this cell.

        This list does not convey any of the CGS Boolean logic

        :rtype: Surfaces
        """
        return self._surfaces

    @property
    def parameters(self):
        """
        A dictionary of the additional parameters for the object.

        e.g.: ``1 0 -1 u=1 imp:n=0.5`` has the parameters
        ``{"U": "1", "IMP:N": "0.5"}``

        :returns: a dictionary of the key-value pairs of the parameters.
        :rytpe: dict
        """
        return self._parameters

    @parameters.setter
    def parameters(self, params):
        if not isinstance(params, dict):
            raise TypeError("parameters must be a dict")
        self._parameters = params

    @property
    def complements(self):
        """
        The Cell objects that this cell is a complement of

        :rytpe: :class:`montepy.cells.Cells`
        """
        return self._complements

    @property
    def cells_complementing_this(self):
        """The cells which are a complement of this cell.

        This returns a generator.

        :rtype: generator
        """
        if self._problem:
            for cell in self._problem.cells:
                if cell != self:
                    if self in cell.complements:
                        yield cell

    def update_pointers(self, cells, materials, surfaces):
        """
        Attaches this object to the appropriate objects for surfaces and materials.

        :param cells: a Cells collection of the cells in the problem.
        :type cells: Cells
        :param materials: a materials collection of the materials in the problem
        :type materials: Materials
        :param surfaces: a surfaces collection of the surfaces in the problem
        :type surfaces: Surfaces
        """
        self._surfaces = Surfaces()
        self._complements = Cells()
        if self.old_mat_number is not None:
            if self.old_mat_number > 0:
                try:
                    self._material = materials[self.old_mat_number]
                except KeyError:
                    raise BrokenObjectLinkError(
                        "Cell", self.number, "Material", self.old_mat_number
                    )
            else:
                self._material = None
        self._geometry.update_pointers(cells, surfaces, self)

    def remove_duplicate_surfaces(self, deleting_dict):
        """Updates old surface numbers to prepare for deleting surfaces.

        :param deleting_dict: a dict of the surfaces to delete.
        :type deleting_dict: dict
        """
        new_deleting_dict = {}
        for dead_surface, new_surface in deleting_dict.items():
            if dead_surface in self.surfaces:
                new_deleting_dict[dead_surface] = new_surface
        if len(new_deleting_dict) > 0:
            self.geometry.remove_duplicate_surfaces(new_deleting_dict)
            for dead_surface in new_deleting_dict:
                self.surfaces.remove(dead_surface)

    def _update_values(self):
        if self.material:
            mat_num = self.material.number
            self._tree["material"]["density"].is_negative = not self.is_atom_dens
        else:
            mat_num = 0
        self._tree["material"]["mat_number"].value = mat_num
        self._geometry._update_values()
        self._tree.nodes["geometry"] = self.geometry.node
        for input_class, (attr, _) in self._INPUTS_TO_PROPERTY.items():
            getattr(self, attr)._update_values()

    def _generate_default_tree(self):
        material = syntax_node.SyntaxNode(
            "material",
            {
                "mat_number": self._generate_default_node(int, 0),
                "density": self._generate_default_node(float, None),
            },
        )
        geom_node = self._generate_default_node(int, -1)
        self._tree = syntax_node.SyntaxNode(
            "cell",
            {
                "cell_num": self._generate_default_node(int, None),
                "material": material,
                "geometry": None,
                "parameters": syntax_node.ParametersNode(),
            },
        )

    def validate(self):
        """
        Validates that the cell is in a usable state.

        :raises: IllegalState if any condition exists that make the object incomplete.
        """
        if self._density and self.material is None:
            raise IllegalState(f"Cell {self.number} has a density set but no material")
        if self.material is not None and not self._density:
            raise IllegalState(
                f"Cell {self.number} has a non-void material but no density"
            )
        if self.geometry is None or len(self.geometry) == 0:
            raise IllegalState(f"Cell {self.number} has no geometry defined.")

    def link_to_problem(self, problem):
        super().link_to_problem(problem)
        self.complements.link_to_problem(problem)
        self.surfaces.link_to_problem(problem)
        for attr, _ in Cell._INPUTS_TO_PROPERTY.values():
            input = getattr(self, attr, None)
            if input:
                input.link_to_problem(problem)

    def __str__(self):
        if self.material:
            mat_num = self.material.number
        else:
            mat_num = 0
        if self._density:
            if self.is_atom_dens:
                units = "atom/b-cm"
            else:
                units = "g/cm3"
            dens_str = f"DENS: {self._density} {units}"
        else:
            dens_str = "DENS: None"
        ret = f"CELL: {self.number}, mat: {mat_num}, {dens_str}"
        if self.universe and self.universe.number != 0:
            ret += f", universe: {self.universe.number}"
        if self.fill.universe:
            ret += f", filled by: {self.fill.universe}"

        return ret

    def __repr__(self):
        ret = f"CELL: {self.number} \n"
        if self.material:
            ret += str(self.material) + "\n"
        else:
            ret += "Void material \n"
        if self._density:
            ret += f"density: {self._density} "
            if self._is_atom_dens:
                ret += "atom/b-cm\n"
            else:
                ret += "g/cc\n"
        for surface in self._surfaces:
            ret += str(surface) + "\n"
        ret += "\n"
        return ret

    def __lt__(self, other):
        return self.number < other.number

    def __invert__(self):
        base_node = UnitHalfSpace(self, True, True)
        return HalfSpace(base_node, Operator.COMPLEMENT)

    def format_for_mcnp_input(self, mcnp_version):
        """
        Creates a string representation of this MCNP_Object that can be
        written to file.

        :param mcnp_version: The tuple for the MCNP version that must be exported to.
        :type mcnp_version: tuple
        :return: a list of strings for the lines that this input will occupy.
        :rtype: list
        """
        self.validate()
        self._update_values()
        self._tree.check_for_graveyard_comments()
        modifier_keywords = {
            cls._class_prefix(): cls for cls in self._INPUTS_TO_PROPERTY.keys()
        }

        def cleanup_last_line(ret):
            last_line = ret.splitlines()[-1]
            # check if adding to end of comment
            if last_line.lower().startswith("c ") and last_line[-1] != "\n":
                return ret + "\n" + " " * BLANK_SPACE_CONTINUE
            if not last_line[-1].isspace():
                return ret + " "
            return ret

        ret = ""
        for key, node in self._tree.nodes.items():
            if key != "parameters":
                ret += node.format()
            else:
                printed_importance = False
                final_param = next(reversed(node.nodes.values()))
                for param in node.nodes.values():
                    if param["classifier"].prefix.value.lower() in modifier_keywords:
                        cls = modifier_keywords[
                            param["classifier"].prefix.value.lower()
                        ]
                        attr, _ = self._INPUTS_TO_PROPERTY[cls]
                        if attr == "_importance":
                            if printed_importance:
                                continue
                            printed_importance = True
                        # add trailing space to comment if necessary
                        ret = cleanup_last_line(ret)
                        ret += "\n".join(
                            getattr(self, attr).format_for_mcnp_input(
                                mcnp_version, param is not final_param
                            )
                        )
                    else:
                        # add trailing space to comment if necessary
                        ret = cleanup_last_line(ret)
                        ret += param.format()
        # check for accidental empty lines from subsequent cell modifiers that didn't print
        ret = "\n".join([l for l in ret.splitlines() if l.strip()])
        return self.wrap_string_for_mcnp(ret, mcnp_version, True)



# Copyright 2024, Battelle Energy Alliance, LLC All Rights Reserved.
__name__ = "montepy.input_parser"
from . import block_type
from . import cell_parser
from . import data_parser
from . import input_reader
from . import material_parser
from . import mcnp_input
from . import parser_base
from . import read_parser
from . import shortcuts
from . import surface_parser
from . import tally_parser
from . import tally_seg_parser
from . import tokens
